@@ -1,22 +1,34 @@
+from six.moves import reload_module
+
+from django.conf import settings
 from django.http import HttpResponseNotFound
 from django.test import TestCase
 from django.test.client import RequestFactory
+from django.contrib.sites.shortcuts import get_current_site
+from django.urls import clear_url_caches
+
 from robustredirects.middleware import RedirectMiddleware
 from robustredirects.models import Redirect
-from django.contrib.sites.models import get_current_site
+
 
 class TestRedirectMiddleWare(TestCase):
     def setUp(self):
         super(TestRedirectMiddleWare, self).setUp()
         # Every test needs access to the request factory.
         self.factory = RequestFactory()
+        # tests should start with an empty url cache
+        clear_url_caches()
+        settings.ROBUST_REDIRECTS_IGNORED_URL_PATHS = None
 
     @staticmethod
     def run_redirect(request):
-        import dynamic_urls
+        from robustredirects import dynamic_urls
 
-        reload(dynamic_urls)
-        middleware = RedirectMiddleware()
+        def get_response(*args, **kwargs):
+            print("Get Response")
+
+        reload_module(dynamic_urls)
+        middleware = RedirectMiddleware(get_response)
         response = HttpResponseNotFound()
         new_response = middleware.process_response(request, response)
         return new_response
@@ -25,20 +37,20 @@ class TestRedirectMiddleWare(TestCase):
         # Create a redirect
         request = self.factory.get('/test/123/')
 
-        redirect = Redirect(from_url='test/(?P<pk>\d+)/', to_url='somethingelse/(?P<pk>\d+)/',
+        redirect = Redirect(from_url=r'test/(?P<pk>\d+)/', to_url=r'somethingelse/(?P<pk>\d+)/',
                             site=get_current_site(request), uses_regex=True)
 
         redirect.save()
         new_response = self.run_redirect(request)
 
         self.assertEqual(new_response.status_code, 301)
-        assert 'somethingelse' in new_response.serialize_headers()
+        self.assertIn(b"somethingelse", new_response.serialize_headers())
 
     def test_redirect_request_gone(self):
         # Create a redirect
         request = self.factory.get('/test/123/')
 
-        redirect = Redirect(from_url='test/(?P<pk>\d+)/', to_url='',
+        redirect = Redirect(from_url=r'test/(?P<pk>\d+)/', to_url='',
                             site=get_current_site(request), uses_regex=True)
 
         redirect.save()
@@ -50,14 +62,14 @@ class TestRedirectMiddleWare(TestCase):
         # Create a redirect
         request = self.factory.get('/test/123/')
 
-        redirect = Redirect(from_url='test/(?P<pk>\d+)/', to_url='somethingelse/(?P<pk>\d+)/',
+        redirect = Redirect(from_url=r'test/(?P<pk>\d+)/', to_url=r'somethingelse/(?P<pk>\d+)/',
                             site=get_current_site(request), http_status=302, uses_regex=True)
 
         redirect.save()
         new_response = self.run_redirect(request)
 
         self.assertEqual(new_response.status_code, 302)
-        assert 'somethingelse' in new_response.serialize_headers()
+        self.assertIn(b"somethingelse", new_response.serialize_headers())
 
     def test_redirect_request_partial_temporary(self):
         # Create a redirect
@@ -70,7 +82,7 @@ class TestRedirectMiddleWare(TestCase):
         new_response = self.run_redirect(request)
 
         self.assertEqual(new_response.status_code, 302)
-        assert 'partialtest' in new_response.serialize_headers()
+        self.assertIn(b"partialtest", new_response.serialize_headers())
 
     def test_redirect_request_partial_permanent(self):
         # Create a redirect
@@ -83,7 +95,26 @@ class TestRedirectMiddleWare(TestCase):
         new_response = self.run_redirect(request)
 
         self.assertEqual(new_response.status_code, 301)
-        assert 'partialtest' in new_response.serialize_headers()
+        self.assertIn(b"partialtest", new_response.serialize_headers())
+
+    def test_redirect_request_two_partial_entries_permanent(self):
+        # Create a redirect
+        old_route = '/invalidroot/partialtest'
+        redirected_route = '/test/partialtest'
+        request = self.factory.get(old_route)
+
+        redirect = Redirect(from_url='/invalidroot', to_url=redirected_route, is_partial=True,
+                            site=get_current_site(request), http_status=301)
+
+        redirect.save()
+        redirect2 = Redirect(from_url=old_route, to_url=redirected_route, is_partial=True,
+                             site=get_current_site(request), http_status=301)
+
+        redirect2.save()
+        new_response = self.run_redirect(request)
+
+        self.assertEqual(new_response.status_code, 301)
+        self.assertEqual(new_response.url, redirected_route)
 
     def test_redirect_request_partial_gone(self):
         # Create a redirect
@@ -108,4 +139,23 @@ class TestRedirectMiddleWare(TestCase):
         new_response = self.run_redirect(request)
 
         self.assertEqual(new_response.status_code, 302)
-        assert '/partialtest/123/' in new_response.serialize_headers()
+        self.assertIn(b"/partialtest/123/", new_response.serialize_headers())
+
+    def test_redirect_exclusion(self):
+        # Create a redirect
+        request = self.factory.get('/api/test/123/')
+
+        settings.ROBUST_REDIRECTS_IGNORED_PREFIXES = '/api'
+
+        redirect = Redirect(from_url='/test/', to_url='partialtest/', is_partial=True,
+                            site=get_current_site(request), http_status=302)
+
+        redirect.save()
+        redirect2 = Redirect(from_url=r'/api/test/(?P<pk>\d+)/', to_url=r'somethingelse/(?P<pk>\d+)/',
+                             site=get_current_site(request), http_status=302, uses_regex=True)
+
+        redirect2.save()
+        new_response = self.run_redirect(request)
+
+        # no redirect should happen
+        self.assertEqual(new_response.status_code, 404)
